from modelon.impact.client.client import Client
from modelon.impact.client.experiment_definition import (
    SimpleFMUExperimentDefinition,
    SimpleModelicaExperimentDefinition,
    Range,
    Choices,
    SimpleExperimentExtension,
)

SimpleExperimentDefinition = SimpleFMUExperimentDefinition  # For compability
